import abc
import functools
import os
import subprocess

import six


class SubprocessError(Exception):
    def __init__(self, cmd, cwd, error):
        super(SubprocessError, self).__init__(
                'Failed to run: "%s" in %s: %s' % (cmd, cwd, error))

class NoTypeError(Exception):
    pass

class UnknownTypeError(Exception):
    pass

class ConfigError(Exception):
    pass

@functools.total_ordering
class Scanner(object, six.with_metaclass(abc.ABCMeta)):
    def __init__(self, name, priority=0):
        """
        @param name         - name of the Scanner.
        @param priority     - Priority for this scanner.  Scanners with a
                              higher priority will be favored for any given
                              path.  If the priority is less than 0, the
                              scanner will be ignored by any automatic
                              scanner picking
        """
        self._name = name
        self._priority = priority


    def __eq__(self, other):
        return (self._name == other._name
                and self._priority == other._priority)

    def __lt__(self, other):
        return self._priority < other._priority

    @abc.abstractmethod
    def is_suitable(self, path):
        """
        Check if this scanner is suitable to run on the given path.

        @param path - path to check
        @return     - True if this scanner is suitable to scan in
                      the specified path
        """
        pass

    @abc.abstractmethod
    def scan(self, path=None, rescan=False):
        """
        Scan for files at the given path.  This assumes that the
        scanner is suitable for scanning (self.is_suitable()).
        If the Scanner is using a cache, then the cache should be
        invalidated and the file list regenerated when rescan is
        True.

        @param path     - path at which to start scanning, if undefined
                          then the current working directory is used
        @param rescan   - force a full rescan of files instead of using
                          a cached list
        @return         - list of detected files
        """
        pass

    def transform(self, path):
        """
        Final tranform for a path.  This can be used to present matches which
        are more user-friendly during the selection process when can later be
        transformed via this method into the required match.

        @param path - path to be transformed
        @return     - the transformed path
        """
        return path


class SimpleScanner(Scanner):
    def __init__(self, name, cmd, priority=0, detect_cmd=None, root_path=None, cache=None):
        """
        Create a scanner.

        @param name         - name of the Scanner.
        @param cmd          - shell command used to scan for files
        @param priority     - Priority for this scanner.  Scanners with a
                              higher priority will be favored for any given
                              path.  If the priority is less than 0, the
                              scanner will be ignored by any automatic
                              scanner picking
        @param detect_cmd   - If specified, this command will be used to
                              check if this scanner can be used for a
                              given path.  If the command returns 0, the
                              scanner will be used
        @param root_path    - If specified, this path serves two purposes:
                              If no detect_cmd is specified, any path that
                              is a child of the root_path will allow the use
                              of this scanner.  Secondly, when scanning, the
                              current working directory will be set to this
                              path
        @param cache        - If specified, path where this scanner will store
                              a cache of files it scans.  By default, calls
                              to scan will just return the files in the cache.
                              This can be changed by passing rescan=True to
                              scan().
        """
        super(SimpleScanner, self).__init__(name, priority)
        self._cmd = cmd
        self._detect_cmd = detect_cmd
        self._root_path = None
        self._cache = None

        if root_path is not None:
            if root_path.startswith('!'):
                try:
                    c = subprocess.Popen(
                            root_path[1:],
                            shell=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
                    stdout, stderr = c.communicate()
                except OSError:
                    raise SubprocessError(root_path[1:], os.path.realpath(os.curdir), stderr)

                if c.returncode != 0:
                    raise SubprocessError(root_path[1:], os.path.realpath(os.curdir), stderr)

                self._root_path = stdout.strip()
            else:
                root_path = os.path.expandvars(root_path)
                root_path = os.path.expanduser(root_path)
                root_path = os.path.normpath(root_path)
                self._root_path = os.path.realpath(root_path)

        if cache is not None:
            cache = os.path.expandvars(cache)
            cache = os.path.expanduser(cache)
            cache = os.path.normpath(cache)
            self._cache = os.path.realpath(cache)

            cachedir = os.path.dirname(os.path.realpath(cache))
            if not os.path.isdir(cachedir):
                os.makedirs(cachedir)

    @classmethod
    def from_configparser(cls, section, parser):
        """
        Create a scanner from a config parser section.

        @param config_section   - config parser object defining a scanner.
        """
        kwds = {}
        if parser.has_option(section, 'detect_cmd'):
            kwds['detect_cmd'] = parser.get(section, 'detect_cmd').replace('\n', ' ')

        if parser.has_option(section, 'root_path'):
            kwds['root_path'] = parser.get(section, 'root_path')

        if parser.has_option(section, 'priority'):
            kwds['priority'] = parser.get(section, 'priority')

        if parser.has_option(section, 'cache'):
            kwds['cache'] = parser.get(section, 'cache')

        cmd = parser.get(section, 'cmd').replace('\n', ' ')

        return cls(section, cmd, **kwds)

    def is_suitable(self, path):
        """
        Check if this scanner is suitable to run on the given path.
        This involves running the detect_cmd to see if we are
        in an appropriate directory type and checking to see if
        the specified path is a descendent of the root_path.

        @param path - path to check
        @return     - True if this scanner is suitable to scan in
                      the specified path
        """
        if self._root_path is not None:
            path = os.path.realpath(os.path.normpath(path))
            if path.startswith(self._root_path):
                return True

        if self._detect_cmd is not None:
            try:
                stderr = ''
                c = subprocess.Popen(
                        self._detect_cmd,
                        cwd=path,
                        shell=True,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE)
                _, stderr = c.communicate()
            except OSError:
                raise SubprocessError(self._detect_cmd, path, stderr)

            if c.returncode == 0:
                return True

        if self._root_path is None and self._detect_cmd is None:
            return True

        return False

    def scan(self, path=None, rescan=False):
        """
        Scan for files at the given path.  This assumes that the
        scanner is suitable for scanning (self.is_suitable()).
        If a root_path was specified for the command, it is used
        as the current working directory for the command, otherwise
        the path itself is.  If a cache was passed to the
        constructor, the file list will be generated by simply
        reading the cache.  This can be bypassed by setting rescan.


        @param path     - path at which to start scanning, if undefined
                          then the current working directory is used
        @param rescan   - force a full rescan of files instead of using
                          a cached list
        @return         - list of detected files
        """
        have_cache = self._cache is not None and os.path.exists(self._cache)
        ret = None

        if have_cache and not rescan:
            with open(self._cache, 'r') as fp:
                ret = [f.strip() for f in fp.read().split()]
        else:
            if path is None:
                path = os.getcwd()

            cwd = path if self._root_path is None else self._root_path

            try:
                stdout = stderr = ''
                c = subprocess.Popen(
                        self._cmd,
                        cwd=cwd,
                        shell=True,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE)
                stdout, stderr = c.communicate()
            except OSError:
                raise SubprocessError(self._cmd, cwd, stderr)

            if c.returncode != 0:
                raise SubprocessError(self._cmd, cwd, stderr)

            ret = [f.strip().decode('UTF-8') for f in stdout.split()]

            if self._cache is not None:
                with open(self._cache, 'w') as fp:
                    fp.write(u'\n'.join(ret))

        return ret

    def transform(self, path):
        if self._root_path:
            root_rel_path = os.path.join(self._root_path, path)
            return os.path.normpath(root_rel_path)
        else:
            return path


class StaticScanner(Scanner):
    """
    This class is used to provided a static list of paths. Another way of
    viewing it is that the paths have been pre-scanned and this scanner simply
    serves the result.
    """

    def __init__(self, paths):
        """
        Creates a StaticScanner

        The StaticScanner contains a list of paths that it provides for
        matching. Thus there is no really scanning and all paths are
        automatically suitable.

        @option paths - a list of paths to use for matching

        """
        self._paths = paths

    def is_suitable(self, path):
        """
        Returns True if the path is suitable

        For the StaticScanner, this function always returns True.

        @option path - the path to evaluate
        @return a boolean indicating whether the path is suitable

        """
        return True

    def scan(self, path=None, rescan=False):
        """
        Scan the path for a list of files.

        @param path     - ignore by the StaticScanner
        @param rescan   - ignore by the StaticScanner
        @return         - list of paths contained in the StaticScanner

        """
        return self._paths


def plugin_scanner_from_configparser(section, parser):
    """
    Create a plugin scanner from a config parser section.  A plugin
    scanner is defined in an external python file.  The configparser
    must have the 'path' and 'object' options defined.  Any other
    options (excluding 'type') are passed as keyword arguments to
    the scanner constructor.  Note that each option value passed in
    this manner will be a string.

    @option type    - this must be set to python
    @option path    - path to the python file containing the scanner
    @option object  - name of the scanner object

    @param section  - section of the config defining a Scanner
    @param parser   - parser contining definition
    @return         - Object derived from the base Scanner class as
                      defined by the config section
    """
    if not parser.has_option(section, 'path'):
        raise ConfigError('path not specified for section "%s"' % (section,))

    if not parser.has_option(section, 'object'):
        raise ConfigError('object not specified for section "%s"' % (section,))

    kwds = {}
    for option in parser.options(section):
        if option in ('type', 'path', 'object'):
            continue
        kwds[option] = parser.get(section, option)

    env = {}
    path = parser.get(section, 'path')
    obj = parser.get(section, 'object')

    try:
        with open(path) as fp:
            exec(compile(fp.read(), path, 'exec'), env)
    except Exception as e:
        raise ConfigError('Failed to load plugin "%s": %s' % (
            path, str(e)))

    try:
        scanner = env[obj](**kwds)
    except Exception as e:
        raise ConfigError('Failed to create %s:%s with args %s: %s' % (
            path, obj, str(kwds), str(e)))

    if not isinstance(scanner, Scanner):
        raise ConfigError('%s:%s is not an instance of fzsl.Scanner' % (
            path, obj))

    return scanner

def scanner_from_configparser(section, parser):
    """
    Create a Scanner from a config parser section.

    @param section  - section of the config defining a Scanner
    @param parser   - parser contining definition
    @return         - Object derived from the base Scanner class as
                      defined by the config section
    """
    if not parser.has_option(section, 'type'):
        raise NoTypeError('type not specified for section "%s"' % (section,))

    scanner_type = parser.get(section, 'type')

    if scanner_type == 'simple':
        scanner = SimpleScanner.from_configparser(section, parser)
    elif scanner_type == 'python':
        scanner = plugin_scanner_from_configparser(section, parser)
    else:
        raise UnknownTypeError('Unknown type "%s" for section "%s"' % (
            scanner_type, section))

    return scanner




import os
import sys
import unittest

TESTDIR = os.path.realpath(os.path.dirname(__file__))
sys.path.insert(0, os.path.realpath(os.path.join(TESTDIR, '..')))

import fzsl

class SimpleScannerTest(unittest.TestCase):
    def setUp(self):
        self.fm = fzsl.FuzzyMatch()

    def test_library(self):
        self.assertEqual(0, self.fm.n_matches)
        self.assertEqual(0, self.fm.n_files)
        self.assertEqual([], self.fm.top_matches())

        with self.assertRaises(KeyError):
            self.fm.score('a')

        self.fm.add_files(['a', 'b', 'c'])
        self.assertEqual(3, self.fm.n_matches)
        self.assertEqual(3, self.fm.n_files)
        self.assertEqual(['a', 'b', 'c'], sorted(self.fm.top_matches()))

        self.assertEqual(0, self.fm.score('a'))
        self.assertEqual(0, self.fm.start('a'))
        self.assertEqual(0, self.fm.end('a'))

        self.fm.reset_files([])
        self.assertEqual(0, self.fm.n_matches)
        self.assertEqual(0, self.fm.n_files)
        self.assertEqual([], self.fm.top_matches())

    def test_update_scores(self):
        files = ['abc/def/ghi', 'abc/d', 'abc/ggg', 'z/1', 'z/2', 'zz/3']

        self.fm.add_files(files)

        self.fm.update_scores("abc")
        self.assertEqual(3, self.fm.n_matches)
        matches = self.fm.top_matches()

        self.fm.update_scores("abcd")
        self.assertEqual(2, self.fm.n_matches)

        self.fm.update_scores("abc")
        self.assertEqual(3, self.fm.n_matches)
        self.assertEqual(matches, self.fm.top_matches())

        self.fm.update_scores("ab")
        self.fm.update_scores("a")
        self.fm.update_scores("")
        self.assertEqual(self.fm.n_matches, self.fm.n_files)

    def test_prefer_latter(self):
        files = ['prefix/abc/stuff', 'abc/d', 'some/prefix/abc']

        self.fm.add_files(files)
        self.fm.update_scores("abc")
        self.assertEqual("some/prefix/abc", self.fm.top_matches(1)[0])

    def test_prefer_shorter(self):
        files = ["a/z/b/z/c", "a/b/c", "a/bbbbb/c"]
        
        self.fm.add_files(files)
        self.fm.update_scores("abc")
        self.assertEqual("a/b/c", self.fm.top_matches(1)[0])

    def test_start_end(self):
        files = ['abc/def/ghi', 'ggg/a/b/ggg/c/d', 'ggg/abc']

        self.fm.add_files(files)
        self.fm.update_scores("abc")

        self.assertEqual(0, self.fm.start('abc/def/ghi'))
        self.assertEqual(2, self.fm.end('abc/def/ghi'))

        self.assertEqual(4, self.fm.start('ggg/a/b/ggg/c/d'))
        self.assertEqual(12, self.fm.end('ggg/a/b/ggg/c/d'))

        self.assertEqual(4, self.fm.start('ggg/abc'))
        self.assertEqual(6, self.fm.end('ggg/abc'))

    def test_ignorecase(self):
        files = ['abc/def/ghi', 'abc/d', 'abc/ggg', 'z/1', 'z/2', 'zz/3']

        self.fm.add_files(files)

        self.fm.update_scores("ABC")
        self.assertEqual(3, self.fm.n_matches)


def main():
    unittest.main()

if __name__ == '__main__':
    main()




import os
import sys
import tempfile
import unittest

try:
    import configparser
    import io
    StringIO = io.StringIO
except ImportError:
    import ConfigParser as configparser
    configparser.RawConfigParser.read_file = configparser.RawConfigParser.readfp
    import cStringIO
    StringIO = cStringIO.StringIO

TESTDIR = os.path.realpath(os.path.dirname(__file__))
sys.path.insert(0, os.path.realpath(os.path.join(TESTDIR, '..')))

import fzsl

class SimpleScannerTest(unittest.TestCase):
    def setUp(self):
        self.bn = os.path.basename(__file__)

    def test_sort(self):
        s1 = fzsl.SimpleScanner('test', 'echo', priority=1)
        s2 = fzsl.SimpleScanner('test', 'echo', priority=2)

        l = [s2, s1]
        l.sort()

        self.assertEqual(l[0], s1)
        self.assertEqual(l[1], s2)

    def test_root_path_match(self):
        s = fzsl.SimpleScanner('test', 'echo', root_path=TESTDIR)
        self.assertTrue(s.is_suitable(TESTDIR))
        self.assertTrue(s.is_suitable('%s/test/stuff' % (TESTDIR,)))
        self.assertFalse(s.is_suitable(os.path.dirname(TESTDIR)))
        self.assertFalse(s.is_suitable('%s/../' % (TESTDIR,)))

    def test_detect_cmd_match(self):
        cmd = '[ -f %s ]' % (self.bn,)
        s = fzsl.SimpleScanner('test', 'echo', detect_cmd=cmd)
        self.assertTrue(s.is_suitable(TESTDIR))
        self.assertFalse(s.is_suitable('%s/../' % (TESTDIR,)))

        cmd = '[ -f thisfileisnothere ]'
        s2 = fzsl.SimpleScanner('test', 'echo', detect_cmd=cmd)
        self.assertFalse(s2.is_suitable(TESTDIR))

    def test_cmd(self):
        cmd = 'find . -name %s' % (self.bn,)
        s = fzsl.SimpleScanner('test', cmd)
        self.assertIn('./' + self.bn, s.scan(TESTDIR))
        self.assertNotIn('.git' + self.bn, s.scan(TESTDIR))

        self.assertEqual(0, len(s.scan('%s/../bin' % (TESTDIR,))))

    def test_fallthrough(self):
        s = fzsl.SimpleScanner('test', 'echo')
        self.assertTrue(s.is_suitable(TESTDIR))

    def test_scanner(self):
        cache = tempfile.NamedTemporaryFile(dir=TESTDIR, mode='w')

        with open(os.path.join(TESTDIR, 'files'), 'r') as src:
            cache.write(u'\n'.join(src.read().split()))
        cache.flush()

        s = fzsl.SimpleScanner('test', 'echo hi', cache=cache.name)
        results = s.scan()
        self.assertEqual(len(results), 49168)

        results = s.scan(rescan=True)
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0], 'hi')

    def test_load_rule(self):
        buf = "[some-rule]\n"

        parser = configparser.RawConfigParser()
        parser.read_file(StringIO(buf))

        with self.assertRaises(fzsl.NoTypeError):
            fzsl.scanner_from_configparser('some-rule', parser)

        buf += "type: junk\n"
        parser.read_file(StringIO(buf))
        with self.assertRaises(fzsl.UnknownTypeError):
            fzsl.scanner_from_configparser('some-rule', parser)

        buf = "[some-rule]\ntype = simple\n"
        parser.read_file(StringIO(buf))
        with self.assertRaises(configparser.NoOptionError):
            fzsl.scanner_from_configparser('some-rule', parser)

        buf += "cmd = echo\n"
        parser.read_file(StringIO(buf))
        r = fzsl.scanner_from_configparser('some-rule', parser)
        self.assertIsInstance(r, fzsl.Scanner)

    def test_load_plugin(self):
        buf = '[rule]\ntype=python\npath=%s/plugins/test_plugin.py\n' % (
                TESTDIR,)

        parser = configparser.RawConfigParser()

        parser.read_file(StringIO(buf + "object=UnsuitableScanner\n"))
        scanner = fzsl.scanner_from_configparser('rule', parser)
        self.assertFalse(scanner.is_suitable(''))

        parser = configparser.RawConfigParser()
        parser.read_file(StringIO(buf + "object=ABCScanner\n"))
        scanner = fzsl.scanner_from_configparser('rule', parser)
        self.assertTrue(scanner.is_suitable(''))
        self.assertEqual(scanner.scan(''), ['a', 'b', 'c'])

        b = buf + "object=KwdsScanner\n"
        b += 'arg1=1\narg2=abc\narg3=some string\n'
        parser = configparser.RawConfigParser()
        parser.read_file(StringIO(b))
        scanner = fzsl.scanner_from_configparser('rule', parser)
        self.assertEqual(scanner.scan(''), ['1', 'abc', 'some string'])

        with self.assertRaises(fzsl.ConfigError):
            parser = configparser.RawConfigParser()
            parser.read_file(StringIO(buf + "object=BrokenScanner1\n"))
            scanner = fzsl.scanner_from_configparser('rule', parser)
                
        with self.assertRaises(fzsl.ConfigError):
            parser = configparser.RawConfigParser()
            parser.read_file(StringIO(buf + "object=BrokenScanner2\n"))
            scanner = fzsl.scanner_from_configparser('rule', parser)

        with self.assertRaises(fzsl.ConfigError):
            parser = configparser.RawConfigParser()
            parser.read_file(StringIO(buf + "object=BrokenScanner3\n"))
            scanner = fzsl.scanner_from_configparser('rule', parser)
 
def main():
    unittest.main()

if __name__ == '__main__':
    main()




import functools
import heapq
import multiprocessing
import re
import signal

class MatchInfo(object):
    def __init__(self, start=0, end=0, score=0, round_ejected=0):
        """
        Information about how a patch matches against the given search

        @attr start         - index in path where match starts
        @attr end           - index in path where match ends
        @attr score         - match score
        @attr round_ejected - the round (length of search string) that this path
                              was removed as a possible match
        """
        self.start = start
        self.end = end
        self.score = score
        self.round_ejected = round_ejected

    def update(self, start=None, end=None, score=None, round_ejected=None):
        """
        Update the MatchInfo attributes

        @param start         - index in path where match starts
        @param end           - index in path where match ends
        @param score         - match score
        @param round_ejected - the round (length of search string) that this path
        """
        if start is not None:
            self.start = start
        if end is not None:
            self.end = end
        if score is not None:
            self.score = score
        if round_ejected is not None:
            self.round_ejected = round_ejected

def default_scorer(path, c_round, regex):
    """
    Score how well a path is matched by the given regex.  Note
    that this function will be passed to a multiprocessing pool
    and needs to return enough context such that the results can
    be correctly collated back at the source.

    @param path     - path to score
    @param c_round  - length of the current search query.  This is
                      used to support deletion of characters in the
                      query.  If a character is deleted, then all
                      results that were ejected in the previous round
                      will be re-evaluated
    @param regex    - regular expression to use for matching
    @return         - This is a little complex in order to support
                      multiprocessing.  The return is a tuple where
                      the first item is the path that was passed into
                      this function.  The second item is another tuple
                      consisting of MatchInfo attributes: start, end,
                      score, round_ejected.  That is:
                      (path, (start, end, score, round_ejected))
    """
    matches = [m for m in regex.finditer(path)]
    if matches:
        def score(match):
            return 1.0 / (len(path) - match.start(1))

        ranked = [(score(m), m) for m in matches]
        score, best = max(ranked)

        return path, (best.start(1), best.end(1) - 1, score, 0)
    else:
        return path, (0, 0, 0.0, c_round)

class FuzzyMatch(object):
    def __init__(self, files=None, scorer=default_scorer):
        """
        Create a FuzzyMatcher which is responsible for handling the
        state as paths are added to the library being searched by an
        updating query.

        @param files    - initial library of paths to rank
        @param scorer   - scoring function.  This function must match
                          the signature and return values of the
                          default_scorer.  See it for more information.
        """
        if files is not None:
            self._library = {path:MatchInfo() for path in files}
        else:
            self._library = {}

        self._scorer = scorer
        self._search = ''

        def pool_init():
            signal.signal(signal.SIGINT, signal.SIG_IGN)

        self._pool = multiprocessing.Pool(initializer=pool_init)

    @property
    def n_matches(self):
        """
        Number of paths which are candidates given the current query
        """
        return len([info for info in self._library.values() if info.round_ejected == 0])

    @property
    def n_files(self):
        """
        Total number of paths in the library being searched
        """
        return len(self._library)

    def add_files(self, files):
        """
        Add files to the library being searched.  This does not automatically
        score the new files.  Use update_scores() to do so if necessary.

        @param files    - list of files to add.
        """
        self._library.update({path:MatchInfo() for path in files})

    def reset_files(self, files):
        """
        Reset the library of possible files to match.  This does not
        automatically score the new files.  Use update_scores() to do
        so if necessary.

        @param files    - new files to use as a library
        """
        self._library = {path:MatchInfo() for path in files}

    def update_scores(self, search):
        """
        Update the scores for every path in the library that was not
        previously ejected.  If the search term is smaller than the
        previous one, paths that were ejected in the last round will
        be re-scored and possibly retained.

        @param search   - query to fuzzy match against the library
        """
        s_len = len(search)

        if s_len < len(self._search):
            _ = [self._library[path].update(round_ejected=0)
                    for path, info in self._library.items()
                    if info.round_ejected == s_len + 1]

        self._search = search

        if s_len == 0:
            return
        elif s_len == 1:
            # In the single character search case, it's far faster to use the
            # string index function over regex.  This is important as we want
            # to slim down the potential matches as quickly as possible.
            #
            # > def f(): 'aoacoeaemaoceaimfoaijhaohfeiahfoefhoeia'.index(j)
            # > def f2(): re.search('j', 'aoacoeaemaoceaimfoaijhaohfeiahfoefhoeia')
            # > timeit.timeit('f()', 'from __main__ import f')
            # 0.5705671310424805
            # > timeit.timeit('f()', 'from __main__ import f2 as f')
            # 2.7029879093170166
            def quick_score(path, info):
                index = path.find(search)
                if index != -1:
                    info.update(start=index, end=index, score=1.0)
                else:
                    info.update(round_ejected=1)

            _ = [quick_score(path, info) for path, info in self._library.items()]
            return

        pattern = '(?=(' + '.*?'.join(re.escape(c) for c in search) + '))'
        regex = re.compile(pattern, re.IGNORECASE)

        scorer = functools.partial(self._scorer, c_round=s_len, regex=regex)
        candidates = [path for path, info
                in self._library.items()
                if info.round_ejected == 0]

        _ = [self._library[path].update(*update)
                for path, update in self._pool.map(scorer, candidates)]

    def score(self, path):
        """
        @param path - path to lookup
        @return     - score of the given path
        """
        return self._library[path].score

    def start(self, path):
        """
        @param path - path to lookup
        @return     - index of the start of the match of the current query in the path
        """
        return self._library[path].start

    def end(self, path):
        """
        @param path - path to lookup
        @return     - index of the end of the match of the current query in the path
        """
        return self._library[path].end

    def top_matches(self, depth=10):
        """
        Get the best matching paths in the library.  Note that only paths
        which have not been ejected and have a positive score will be returned.
        Therefore, the length of the returned list may be less than the
        specified depth.

        @param depth    - maximum number of paths to return
        @return         - sorted list of the top scoring paths in the library
        """
        if len(self._search) > 0:
            valid = [path
                    for path, info in self._library.items()
                    if info.score > 0 and info.round_ejected == 0]
        else:
            valid = self._library.keys()

        ret = heapq.nlargest(depth, valid, key=lambda x: self._library[x].score)
        return ret




from . core import (
    FuzzyMatch,
)

from . ui import (
    ncurses,
    SimplePager,
)

from . scanner import (
    ConfigError,
    NoTypeError,
    Scanner,
    scanner_from_configparser,
    SimpleScanner,
    StaticScanner,
    SubprocessError,
    UnknownTypeError,
)




#!/usr/bin/env python
import os
import setuptools

def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()

setuptools.setup(
        name='fzsl',
        version='0.3',
        description='Fuzzy path searching for shells',
        license='BSD',
        long_description=(read('README.rst')),
        author='Justin Bronder',
        author_email='jsbronder@gmail.com',
        url='http://github.com/jsbronder/fzsl',
        keywords='fuzzy shell search console match ncurses',
        packages=['fzsl'],
        package_data = {
          '': ['*.rst', 'etc/fzsl.*', 'LICENSE'],
        },
        data_files = [
          ('share/fzsl/', ['README.rst', 'LICENSE', 'etc/fzsl.bash', 'etc/fzsl.conf']),
        ],
        scripts=['bin/fzsl'],
        install_requires=['six'],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Environment :: Console :: Curses',
            'Intended Audience :: Developers',
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: BSD License',
            'Operating System :: POSIX',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Topic :: System :: Shells',
            'Topic :: System :: Systems Administration',
            'Topic :: System :: System Shells',
            'Topic :: Terminals',
            'Topic :: Terminals :: Serial',
            'Topic :: Terminals :: Telnet',
            'Topic :: Terminals :: Terminal Emulators/X Terminals',
            'Topic :: Text Processing :: Filters',
            'Topic :: Utilities',
        ]
)



import contextlib
import curses
import os
import sys

import six

import fzsl

for i, color in enumerate(('white', 'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan')):
    vars()['COL_%s' % (color.upper())] = i
    vars()['COL_B%s' % (color.upper())] = i + 8


@contextlib.contextmanager
def ncurses():
    """
    Context manager for curses applications.  This does a number of
    things beyond what curses.wrapper() does.

    - Redirect stdout to stderr.  This is done so that we can still
      use the ncurses interface from within a pipe or subshell.
    - Drop the escape delay down to 25ms, similar to vim.
    - Remove newline mode.

    An ncurses screen is returned by the manager.  If any exceptions
    occur, all of the setup performed by the manager is undone before
    raising the original exception.  This should guarantee that any
    bugs in the code will not leave the user with a messed up shell.
    """
    # Push stdout to stderr so that we still get the curses
    # output while inside of a pipe or subshell
    old_stdout = sys.__stdout__
    old_stdout_fd = os.dup(sys.stdout.fileno())
    os.dup2(sys.stderr.fileno(), sys.stdout.fileno())

    # Reduce the timeout after receiving an escape character, there
    # doesn't seem to be a way to do this via curses so we have to
    # set the environment variable before creating the screen.
    if not 'ESCDELAY' in os.environ:
        os.environ['ESCDELAY'] =  '25'

    scr = curses.initscr()
    curses.start_color()
    curses.use_default_colors()
    _ = [curses.init_pair(i + 1, i, -1) for i in range(curses.COLORS)]

    curses.noecho()
    curses.cbreak()
    curses.raw()
    curses.nonl()
    curses.curs_set(2)
    scr.keypad(1)

    exc = None
    try:
        yield scr
    #pylint: disable=W0703
    except Exception:
        exc = sys.exc_info()

    scr.keypad(0)
    curses.nl()
    curses.nocbreak()
    curses.echo()
    curses.endwin()

    os.dup2(old_stdout_fd, sys.stdout.fileno())
    sys.stdout = old_stdout

    if exc is not None:
        six.reraise(exc[0], exc[1], exc[2])


class SimplePager(object):
    def __init__(self, scr, scanner):
        """
        Create a simple pager for showing scan results.  As new terms
        are entered, the results will be updated with the best matches
        showing a highlight over the part of the string which matched.

        @param scr      - ncurses screen to use
        @param scanner  - fzsl.Scanner object that will be used to
                          generate the list of possible matches
        """
        self._scr = scr
        self._scanner = scanner

        self._show_score = False
        self._fm = fzsl.FuzzyMatch()
        self._selection = 0
        self._search = ''

        y, x = self._scr.getmaxyx()

        self._prompt = curses.newwin(1, x, y-1, 0)
        self._select = curses.newwin(y - 2, x, 0, 0)
        self._max_y = y - 2
        self._max_x = x - 1
        self._cursor_x = 0

    def _draw_select(self):
        """
        Redraw the selection window which contains all of the
        possible matches to the current search.
        """
        self._select.erase()
        m = self._fm.top_matches(self._max_y)
        if self._selection >= len(m):
            self._selection = max(len(m) - 1, 0)

        for index, match in enumerate(m):
            if len(self._search) > 0 and self._fm.score(match) == 0:
                continue

            prefix = u''
            if self._show_score:
                prefix = u'%f     ' % (self._fm.score(match),)
            offset = len(prefix)

            start = self._fm.start(match)
            end = self._fm.end(match)
            if end > 0 and len(self._search) > 0:
                end += 1
            line = self._max_y - index - 1
            decor = 0
            if self._selection == index:
                decor = curses.A_UNDERLINE

            match = match[:self._max_x]
            start = min(start, self._max_x)
            end = min(end, self._max_x)

            self._select.addstr(line, 0, prefix + match[:start], decor)
            if start + offset < self._max_x:
                self._select.addstr(line, start+offset, match[start:end], decor|curses.color_pair(COL_BCYAN))
            if end + offset < self._max_x:
                self._select.addstr(line, end+offset, match[end:], decor)
        self._select.refresh()

    def _draw_prompt(self):
        """
        Redraw the prompt window.
        """
        self._prompt.erase()
        prompt = "%d/%d >" % (self._fm.n_matches, self._fm.n_files)
        search_start = 4 + len(prompt)

        self._prompt.addstr(0, 2, prompt)
        self._prompt.addstr(0, search_start, self._search)
        y, x = self._prompt.getyx()
        self._prompt.move(y, search_start + self._cursor_x)
        self._prompt.refresh()

    def _scan(self, rescan=False):
        return [f.encode('unicode-escape') for f in self._scanner.scan(rescan=rescan)]

    def run(self):
        """
        Start the pager.
        """
        self._scr.addstr("Scanning ...")
        self._scr.refresh()
        files = self._scan()
        self._fm.add_files(files)

        self._draw_select()
        self._draw_prompt()

        while True:
            c = self._scr.getch()
            key = curses.keyname(c).decode('UTF-8')

            if key in (u'^M',):
                # enter
                break
            elif key  in ('KEY_DOWN', '^J'):
                # down arrow, ctrl+j
                self._selection = self._selection - 1 if self._selection > 0 else 0
                self._draw_select()
            elif key  in ('KEY_UP', '^K'):
                # up arrow, ctrl+k
                self._selection = self._selection + 1 if self._selection < self._max_y - 2 else self._selection
                self._draw_select()
            elif key in ('KEY_LEFT',):
                if self._cursor_x > 0:
                    self._cursor_x -= 1
                    self._draw_prompt()
            elif key in ('KEY_RIGHT',):
                if self._cursor_x < len(self._search):
                    self._cursor_x += 1
                    self._draw_prompt()
            elif key in ('^V',):
                # ctrl+v
                self._show_score = not self._show_score
                self._draw_select()
            elif key in ('^[', '^C'):
                # escape or ctrl+c
                return ''
            elif c in (curses.KEY_RESIZE,):
                y, x = self._scr.getmaxyx()

                # Have to handle the parent screen fully first otherwise updates
                # to the subwindows don't seem to take.
                curses.resizeterm(y, x)
                self._scr.resize(y, x)
                self._scr.erase()
                self._scr.refresh()

                self._max_y = y - 2
                self._max_x = x - 1
                self._cursor_x = x if x < self._cursor_x else self._cursor_x

                self._select.resize(y - 2, x)
                self._draw_select()

                self._prompt.resize(1, x)
                self._prompt.mvwin(y-1, 0)
                self._draw_prompt()
            else:
                if key in ('KEY_BACKSPACE',):
                    # delete, backspace
                    if self._cursor_x > 0:
                        start = self._search[:self._cursor_x - 1]
                        end = self._search[self._cursor_x:]
                        self._search = start + end
                        self._cursor_x -= 1
                elif key in ('^R', 'KEY_F(5)'):
                    self._scr.erase()
                    self._scr.addstr('Scanning ...')
                    self._scr.refresh()
                    files = self._scan(rescan=True)
                    self._fm.reset_files(files)
                else:
                    start = self._search[:self._cursor_x]
                    end = self._search[self._cursor_x:]
                    self._search = start + chr(c) + end
                    self._cursor_x += 1

                self._fm.update_scores(self._search)

                self._draw_select()
                self._draw_prompt()

        try:
            match = self._fm.top_matches(self._max_y)[self._selection]
            return self._scanner.transform(match)
        except IndexError:
            return ''

